#define __LKT_MODULE_MAIN_SOURCE__
#include "module_qt_window.hh"

#include <QApplication>

#include <lektor/lktmodule.h>
#include <sched.h>
#include "mpv.h"

#include "qt_window/mainwindow.hh"

/************************
 * Function definitions *
 ************************/

static void module_qt_window_free(module_qt_window_s *);
static void module_qt_window_close(module_qt_window_s *);
static bool module_qt_window_new(module_qt_window_s **, queue *, lkt_db *);

/*********************
 * Private functions *
 *********************/

/********************************
 * va_list version of functions *
 ********************************/

static int
mod_new(va_list *va)
{
    va_list copy;
    va_copy(copy, *va);
    module_qt_window_s **win = reinterpret_cast<module_qt_window_s **>(va_arg(copy, void **));
    queue *queue             = va_arg(copy, struct queue *);
    lkt_db *db               = va_arg(copy, lkt_db *);
    bool ret                 = module_qt_window_new(win, queue, db);
    va_end(copy);
    return !ret;
}

static int
mod_close(va_list *va)
{
    va_list copy;
    va_copy(copy, *va);
    module_qt_window_s **win = reinterpret_cast<module_qt_window_s **>(va_arg(copy, void **));
    module_qt_window_close(*win);
    va_end(copy);
    return 0;
}

static int
mod_free(va_list *va)
{
    va_list copy;
    va_copy(copy, *va);
    module_qt_window_s **win = reinterpret_cast<module_qt_window_s **>(va_arg(copy, void **));
    module_qt_window_free(*win);
    va_end(copy);
    return 0;
}

static int
mod_toggle_pause(va_list *va)
{
    va_list copy;
    va_copy(copy, *va);
    module_qt_window_s **win = reinterpret_cast<module_qt_window_s **>(va_arg(copy, void **));
    bool ret                 = (*win)->mpv_widget->toggle_pause();
    va_end(copy);
    return !ret;
}

static int
mod_load_file(va_list *va)
{
    va_list copy;
    va_copy(copy, *va);
    module_qt_window_s **win = reinterpret_cast<module_qt_window_s **>(va_arg(copy, void **));
    const char *file         = va_arg(copy, const char *);
    const bool ret           = (*win)->mpv_widget->load_file(file);
    va_end(copy);
    return !ret;
}

static int
mod_set_volume(va_list *va)
{
    va_list copy;
    va_copy(copy, *va);
    module_qt_window_s **win = reinterpret_cast<module_qt_window_s **>(va_arg(copy, void **));
    const int volume         = va_arg(copy, int);
    const bool ret           = (*win)->mpv_widget->set_volume(volume);
    va_end(copy);
    return !ret;
}

static int
mod_set_paused(va_list *va)
{
    va_list copy;
    va_copy(copy, *va);
    module_qt_window_s **win = reinterpret_cast<module_qt_window_s **>(va_arg(copy, void **));
    const int state          = va_arg(copy, int);
    const bool ret           = (*win)->mpv_widget->set_paused(state);
    va_end(copy);
    return !ret;
}

static int
mod_set_position(va_list *va)
{
    va_list copy;
    va_copy(copy, *va);
    module_qt_window_s **win = reinterpret_cast<module_qt_window_s **>(va_arg(copy, void **));
    const int seconds        = va_arg(copy, int);
    const bool ret           = (*win)->mpv_widget->set_position(seconds);
    va_end(copy);
    return !ret;
}

static int
mod_get_duration(va_list *va)
{
    va_list copy;
    va_copy(copy, *va);
    module_qt_window_s **win = reinterpret_cast<module_qt_window_s **>(va_arg(copy, void **));
    int *duration            = va_arg(copy, int *);
    bool ret                 = (*win)->mpv_widget->get_duration(duration);
    va_end(copy);
    return !ret;
}

static int
mod_get_elapsed(va_list *va)
{
    va_list copy;
    module_qt_window_s **win;
    va_copy(copy, *va);
    win            = reinterpret_cast<module_qt_window_s **>(va_arg(copy, void **));
    int *elapsed   = va_arg(copy, int *);
    const bool ret = (*win)->mpv_widget->get_elapsed(elapsed);
    va_end(copy);
    return !ret;
}

/********************
 * The module stuff *
 ********************/

extern "C" {
REG_BEGIN(qt_window_reg)
REG_ADD_NAMED("new", mod_new)
REG_ADD_NAMED("free", mod_free)
REG_ADD_NAMED("close", mod_close)
REG_ADD_NAMED("toggle", mod_toggle_pause)
REG_ADD_NAMED("load", mod_load_file)
REG_ADD_NAMED("set_paused", mod_set_paused)
REG_ADD_NAMED("set_position", mod_set_position)
REG_ADD_NAMED("set_volume", mod_set_volume)
REG_ADD_NAMED("get_duration", mod_get_duration)
REG_ADD_NAMED("get_elapsed", mod_get_elapsed)
REG_END()
#if !defined(LKT_STATIC_MODULE)
REG_EXPORT(qt_window_reg)
#endif
}

/****************************
 * Private helper functions *
 ****************************/

static void *
___create_mpv_widget(void *arg)
{
    lkt_thread_set_name("lektord/qt-win");

    /* Unsage reinterpret cast, but isok */
    module_qt_window_s *qt_window = reinterpret_cast<module_qt_window_s *>(arg);

    int argc = 0;
    QApplication app(argc, nullptr);
    setlocale(LC_NUMERIC, "C");
    qt_window->main_window = new MainWindow(qt_window);
    qt_window->main_window->show();
    app.exec();

    return nullptr;
}

/***************************
 * Function implementation *
 ***************************/

static bool
module_qt_window_new(module_qt_window_s **win, queue *queue, lkt_db *db)
{
    (void)mod_new;
    (void)mod_free;
    (void)mod_close;
    RETURN_UNLESS(win, "Invalid arguments", false);

    if (*win == nullptr) {
        *win = LKT_ALLOC_STRUCT(module_qt_window_s);
        RETURN_UNLESS(*win, "Out of memory", false);
        memset(*win, 0, sizeof(module_qt_window_s));

        (*win)->queue    = queue;
        (*win)->db       = db;
        (*win)->launched = false;

        pthread_t thread;
        RETURN_IF(pthread_create(&thread, nullptr, ___create_mpv_widget, *win),
                  "Failed to launch the Qt thread", false);

        timespec timeToWait = { 0, 10000000 };
        while (!(*win)->launched)
            nanosleep(&timeToWait, nullptr);
    }

    return true;
}

static void
module_qt_window_close(module_qt_window_s *win)
{
    RETURN_UNLESS(win && win->mpv_widget && win->main_window, "Invalid arguments", NOTHING);
    win->mpv_widget->stop();
}

static void
module_qt_window_free(module_qt_window_s *win)
{
    RETURN_UNLESS(win, "Invalid arguments", NOTHING);
    module_qt_window_close(win);
    win->main_window->closeMpv();
}
